!+
! Module: ebadpixupdate_mod
!
! Authors:  J. Ballet - SAp  CEA Saclay
!           N. Webb - CESR, Toulouse
!          
! Description: Update event list with new bad pixels from CCFs or files
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module ebadpixupdate_mod

  use emutils_mod
  use badpixutils_mod
  use cal_mos
  use eventAttributes
  use datafilter  
  use dal
  use param
  use cal
  use errorHandling

  implicit none 

  private

! Global variables

  logical :: forcereplace

  public ebadpixupdate

contains

  subroutine ebadpixupdate

    character(len=1024) :: event_in, badpix_in
    character(len=8)    :: instrument
    type(DataSetT)      :: ev_set, bad_set
    type(TableT)        :: ev_tab, bad_tab
    integer(int8), dimension(:), pointer :: listofccds
    character(len=1024), dimension(:), pointer :: listofbadpixsets
    integer(kind=int16), allocatable, dimension(:) :: badx, bady, badyext
    integer(kind=int16), allocatable, dimension(:) :: badtype, badflag
    integer             :: nbad, icount, NumBadPixExt, maxnodenumber, node_id
    integer             :: numofccds, jcount, numofbadpixsets, maxccdnumber
    integer(kind=int32) :: calIval
    logical             :: fromccf, fromfiles, overwrite, replaceonboard
    logical, dimension(19) :: ovwccd, obrepccd
  
! Keywords to check for compatibility of badpix and event files

    character(len=8), dimension(2), parameter :: keycompat = &
         (/"TELESCOP","INSTRUME"/)

!  These are the parameters put in by the user

    event_in  = stringParameter("eventset")
    fromccf   = booleanParameter("fromccf")
    fromfiles = booleanParameter("fromfiles")
    ovwccd = booleanParameter("overwrite")
    forcereplace   = booleanParameter("forcereplace")
    obrepccd = booleanParameter("replaceonboard")

    if (.not. fromccf .and. .not. fromfiles)  call message(AppMsg,Sparse,  &
           "fromccf and fromfiles both unset. Nothing done") 

!  Read the input file

    ev_set = dataSet(event_in, Modify)
    ev_tab = table(ev_set, "EVENTS")


! Updating for the CCF origin
! Set ovwccd and obrepccd to False when already done

    ccfpath: if (fromccf) then
       call CAL_setState(block(ev_tab))
       call CAL_getState(instrument=calIval)
       write (errstr,*) "Instrument = ",instId(calIval)
       call message(AppMsg,Verbose)
       instrument = instId(calIval)

!  Get the maximum number of CCDs and nodes for the camera 

       if (index(instrument,'EMOS') > 0) then
          maxccdnumber = 7
          maxnodenumber = 1
       else if (trim(instrument) .eq. 'EPN') then
          maxccdnumber = 12 
          maxnodenumber = 0
       else
          call error('InstrumentNotValid',  &
                     'Instrument must be EMOS1, EMOS2 or EPN')
       end if

       numofccds = parameterCount("ccds")
       allocate(listofccds(1:numofccds))
       do icount=1,numofccds
          listofccds(icount)=intParameter("ccds",icount-1)
       end do

!   Selecting which CCDs to update

       selectallccds: if (numofccds == 1 .and. listofccds(1) == 0) then
          do jcount=1,maxccdnumber
             do node_id=0,maxnodenumber
                NumBadPixExt=jcount+10*node_id
                call loopoverallccds(NumBadPixExt,ev_set,ev_tab,instrument,  &
                     ovwccd(NumBadPixExt), obrepccd(NumBadPixExt), nbad)
                ovwccd(NumBadPixExt) = .false.
                obrepccd(NumBadPixExt) = .false.
             end do
          end do

! When the user entered a list of CCDs

       else
          do jcount=1,numofccds
            NumBadPixExt=listofccds(jcount)
            call loopoverallccds(NumBadPixExt, ev_set, ev_tab, instrument,  &
                 ovwccd(NumBadPixExt), obrepccd(NumBadPixExt), nbad)
            ovwccd(NumBadPixExt) = .false.
            obrepccd(NumBadPixExt) = .false.
          end do
       end if selectallccds
       deallocate(listofccds)
    end if ccfpath

!  Update from a files origin

    filespath: if (fromfiles) then
       numofbadpixsets = parameterCount("badpixtables")
       allocate(listofbadpixsets(1:numofbadpixsets))
       do icount=1,numofbadpixsets
          listofbadpixsets(icount)=stringParameter("badpixtables",icount-1)
       end do

!  Loop over all the files

       do icount=1,numofbadpixsets
          badpix_in = listofbadpixsets(icount)
          call testName(badpix_in)     ! Fill in if set instead of table
          bad_set = dataSet(badpix_in, Read)
          bad_tab = table(bad_set, badpix_in)
          if (equalKeywords(attributable(bad_tab),  &
                            attributable(ev_tab), keycompat)) then
            write(errstr,*) "Reading bad pixels from ", trim(badpix_in)
            call message(AppMsg,Noisy)
            nbad = numberOfRows(bad_tab)
            numBadPixExt = getCcd(block(bad_tab))
            allocate(badx(2*nbad), bady(2*nbad), badyext(2*nbad))
            allocate(badtype(2*nbad), badflag(2*nbad))
            call readBadpix(bad_tab, .true., badx, bady, badyext, badtype,  &
                            badflag, nbad)

            overwrite = ovwccd(NumBadPixExt)
            replaceonboard = obrepccd(NumBadPixExt)
            if (evflagImage(ev_tab, NumBadPixExt, badx, bady, badyext,  &
                 badtype, badflag, nbad, overwrite, replaceonboard))  &
              call addToBadPixExt(ev_set, NumBadPixExt, overwrite,  &
                 replaceonboard, badx, bady, badyext, badtype, badflag, nbad)
            ovwccd(NumBadPixExt) = .false.
            obrepccd(NumBadPixExt) = .false.
            deallocate(badx, bady, badyext, badtype, badflag)
          endif
          call release(bad_set)
       end do
       deallocate(listofbadpixsets)
    end if filespath

    call release(ev_set)

  end subroutine ebadpixupdate

! Flag events on and next to bad pixels
! If overwrite=Y, first remove previous such flags
! Returns False if no event found for that CCD, True otherwise

  logical function evflagImage(ev_tab, ccdnr, badx, bady, badyext, badtype,  &
                               badflag, nbad, overwrite, replaceonboard)

! ev_tab  : Handle to the input/output event table
! ccdnr   : CCD/node number (like CCDNR column)
! badx    : array of RAWX coordinates
! bady    : array of RAWY coordinates
! badyext : array of RAWY extensions
! badtype : array of bad pixel types
! badflag : array of bad pixel status (uplinked, CCF or new)
! nbad    : number of bad pixels
! overwrite : if true, new bad pixels replace old ones
! replaceonboard : if true, new on-board bright pixels replace old ones

    type(TableT), intent(in) :: ev_tab
    integer,      intent(in) :: ccdnr, nbad
    integer(kind=int16), dimension(:), intent(in) :: badx, bady, badyext, &
                                                     badtype, badflag
    logical,   intent(inout) :: overwrite, replaceonboard

    integer(kind=int16), dimension(:), pointer :: posx, posy
    integer(kind=int32), dimension(:), pointer :: quality, row1, row2
    integer(kind=int32), dimension(-2:EMOS_MAX_X+2,-1:EMOS_MAX_Y+2) :: flag_map
    character(len=1024) :: expression
    character(len=8)    :: ExtName, string
    type(ColumnT)       :: flag_col, x_col, y_col
    integer(int32)      :: curflag, testflag
    integer             :: irow, iint, ibad, nmod, nonbad, nflagged, nflag0
    integer             :: xcur, ycur, ycur2, nrows, nblks, ir1, ir2

    call message(AppMsg,Noisy,"Entering EVFLAG")

    evflagImage = .true.

! Select on CCDNR and check event position is acceptable

    write(expression,  &
          '("(CCDNR==",I2,") && (RAWX in 1:",I3,") && (RAWY in 1:",I3,")")')  &
          ccdnr, EMOS_MAX_X, EMOS_MAX_Y
    write(errstr,*) "Call initFilter with expression ", expression
    call message(AppMsg,Noisy)
    call initFilter( expression )

! Get list of intervals where condition is verified (normally only one)

    nrows = doSelect(ev_tab, row1, row2)
    nblks = size(row1)

! Bail out immediately if no event was selected

    if (nblks == 0) then 

! Check whether the bad pixels extension exists already
! If so, prevent overwrite unless forcereplace

       write(string,'(I2.2)') ccdnr
       ExtName = 'BADPIX' // string
       if (hasBlock(parent(ev_tab),ExtName)) then
          if ((overwrite .or. replaceonboard) .and. .not. forcereplace) then
             write(errstr,*) "No event for CCD ", ccdnr,  &
                             ". Switching to incremental mode."
             call warning("evflagImage14")
             overwrite = .false.
             replaceonboard = .false.
          endif
       else   ! Means that this CCD is absent from the events list
          evflagImage = .false.
       endif

! No need to do the following in that case

       call selectReleaseMemory(row1, row2)
       return
    endif

! flag_map is the map of flags

    flag_map = 0

! Project the bad pixels onto it
! Type: 0=OK, 1=hot, 2=flickering, 3=dead, 4=pinhole, 5=anomalous
! Flag: 1=uplinked, 2=in CCF but not uplinked, 3=new

    do ibad = 1,nbad
       xcur = badx(ibad)
       ycur = bady(ibad)
       ycur2 = ycur + badyext(ibad) - 1

! Check that no bad pixel has invalid coordinates (RAWX=0 is valid)

       if (min(xcur+1,ycur) < 1 .or.  &
           min(EMOS_MAX_X-xcur,EMOS_MAX_Y-ycur2) < 0) then
          write(errstr,*) "invalid bad pixel coordinates RAWX =", xcur, &
                      ", RAWY =", ycur, ", YEXTENT =", ycur2, " at row", ibad
          call warning("evflagImage13")
       else
          select case(int(badtype(ibad),kind=int32))
          case (INTACT)
             curflag = 0
          case (DEAD)
             curflag = EVATT_CLOSE_TO_DEADPIX
             flag_map(xcur,ycur:ycur2) = ior(flag_map(xcur,ycur:ycur2),  &
                                             EVATT_ON_BADPIX)
          case default
             if (badflag(ibad) > 1) then
                curflag = EVATT_CLOSE_TO_BRIGHTPIX
             else
                curflag = EVATT_CLOSE_TO_ONBOARD_BADPIX
             endif
             flag_map(xcur,ycur:ycur2) = ior(flag_map(xcur,ycur:ycur2),  &
                                             EVATT_ON_BADPIX)
          end select

! Only pixels adjacent (not diagonal) to bad pixels are flagged

          flag_map(xcur,ycur-1)  = ior(flag_map(xcur,ycur-1),curflag)
          flag_map(xcur,ycur2+1) = ior(flag_map(xcur,ycur2+1),curflag)
          flag_map(xcur-1,ycur:ycur2) = ior(flag_map(xcur-1,ycur:ycur2), &
                                            curflag)
          flag_map(xcur+1,ycur:ycur2) = ior(flag_map(xcur+1,ycur:ycur2), &
                                            curflag)
       endif
    enddo
    call setEventAttribute(ev_tab,EVATT_ON_BADPIX)
    call setEventAttribute(ev_tab,EVATT_CLOSE_TO_DEADPIX)
    call setEventAttribute(ev_tab,EVATT_CLOSE_TO_BRIGHTPIX)
    call setEventAttribute(ev_tab,EVATT_CLOSE_TO_ONBOARD_BADPIX)

! Read the columns (could be done once and for all in separate routine)

    flag_col = column(ev_tab, EVATT_COLUMN_NAME, Modify)
    quality => int32Data(flag_col)
    x_col = column(ev_tab, "RAWX", Read)
    posx => int16Data(x_col)
    y_col = column(ev_tab, "RAWY", Read)
    posy => int16Data(y_col)

! If events flagged as bad exist and overwrite, reset everything first

    nflagged = 0
    nonbad = 0
    do iint = 0,nblks-1
       ir1 = row1(iint)
       ir2 = row2(iint)
       nflagged = nflagged + count(quality(ir1:ir2) > 0)
       nonbad = nonbad  &
              + sum(iand(quality(ir1:ir2),EVATT_ON_BADPIX)/EVATT_ON_BADPIX)
    enddo

    if (overwrite) then
       if (nonbad > 0 .or. forcereplace) then
          testflag = EVATT_ON_BADPIX + EVATT_CLOSE_TO_DEADPIX +  &
                     EVATT_CLOSE_TO_BRIGHTPIX
          testflag = NOT(testflag)
          do iint = 0,nblks-1
             ir1 = row1(iint)
             ir2 = row2(iint)
             quality(ir1:ir2) = iand(quality(ir1:ir2),testflag)
          enddo
          write(errstr,*) nonbad, " events cleared of ON_BADPIX for CCDNR =", &
                          ccdnr
          call message(AppMsg,Verbose)
          nonbad = 0
       else
          write(errstr,*) "No event flagged as ON_BADPIX for CCD ", ccdnr,  &
                          ". This probably means that the bad pixels were ",  &
                          "taken out already. Switching to incremental mode."
          call warning("evflagImage12")
          overwrite = .false.
       endif
    endif

! If flagged events exist and replaceonboard, reset everything first

    if (replaceonboard) then
       if (nflagged > 0 .or. forcereplace) then
          nflagged = 0
          testflag = NOT(EVATT_CLOSE_TO_ONBOARD_BADPIX)
          do iint = 0,nblks-1
             ir1 = row1(iint)
             ir2 = row2(iint)
             nflag0 = sum(iand(quality(ir1:ir2),EVATT_CLOSE_TO_ONBOARD_BADPIX)&
                          /EVATT_CLOSE_TO_ONBOARD_BADPIX)
             if (nflag0 > 0) then
                quality(ir1:ir2) = iand(quality(ir1:ir2),testflag)
                nflagged = nflagged + nflag0
             endif
          enddo
          if (nflagged > 0) then
             write(errstr,*) nflagged, " events cleared of ",  &
                  "CLOSE_TO_ONBOARD_BADPIX for CCDNR =", ccdnr
             call message(AppMsg,Verbose)
          endif
       else
          write(errstr,*) "No event flagged at all for CCD ", ccdnr,  &
               ". This probably means that the flagged events were ",  &
               "taken out already. Switching to incremental mode."
          call warning("evflagImage11")
          replaceonboard = .false.
       endif
    endif

    nmod = 0
    testflag = EVATT_CLOSE_TO_DEADPIX + EVATT_CLOSE_TO_BRIGHTPIX  &
             + EVATT_CLOSE_TO_ONBOARD_BADPIX
    do iint = 0,nblks-1
       do irow = row1(iint),row2(iint)
          xcur = posx(irow)
          ycur = posy(irow)
          curflag = quality(irow)
          quality(irow) = ior(curflag,flag_map(xcur,ycur))

! Check whether event is newly flagged as close to badpix

          if (iand(quality(irow)-curflag,testflag) > 0)  nmod = nmod + 1
       enddo
    enddo

    nonbad = - nonbad
    do iint = 0,nblks-1
       ir1 = row1(iint)
       ir2 = row2(iint)
       nonbad = nonbad  &
              + sum(iand(quality(ir1:ir2),EVATT_ON_BADPIX)/EVATT_ON_BADPIX)
    enddo
    write(errstr,*) nonbad, " events newly flagged as ON_BADPIX for CCDNR =", &
                    ccdnr
    call message(AppMsg,Verbose)

    write(errstr,*) nmod, " events now flagged as CLOSE_TO_BADPIX"
    call message(AppMsg,Verbose)

! Release the columns

    call release(x_col)
    call release(y_col)
    call release(flag_col)

! Release the selectlib pointers

    call selectReleaseMemory(row1, row2)
    call message(AppMsg,Noisy,"Exiting EVFLAG")

  end function evflagImage


  subroutine addToBadPixExt(ev_set, NumBadPixExt, overwrite, replaceonboard,  &
                            badx, bady, badyext, badtype, badflag, nbad)

!  Write the bad pixels into the bad pixel extension

    type(DataSetT), intent(in) :: ev_set    
    integer,        intent(in) :: NumBadPixExt
    logical,        intent(in) :: overwrite, replaceonboard
    integer(kind=int16), dimension(:), intent(inout) :: badx, bady, badyext
    integer(kind=int16), dimension(:), intent(inout) :: badtype, badflag
    integer,                           intent(inout) :: nbad

    character(len=8)         :: ExtName, string
    type(TableT)             :: evbadpix_tab, tab_bpix 
    type(ColumnT)            :: badx_col,bady_col,badt_col,bade_col,badf_col
    integer(kind=int16), allocatable, dimension(:) :: bad2x, bad2y, bad2ext
    integer(kind=int16), allocatable, dimension(:) :: bad2type, bad2flag
    integer(kind=int16), dimension(:), pointer :: bad1x, bad1y, bad1type
    integer(kind=int16), dimension(:), pointer ::  bad1ext, bad1flag
    integer                  :: tablength, tablength2, k
    integer                  :: newnumber, anothernumber

    call message(AppMsg,Noisy,"Getting bad pixel tables")

    write(string,'(I2.2)') NumBadPixExt
    ExtName='BADPIX'//string

!    If the extension exists already, add the new bad pixels to those known

    if (hasBlock(ev_set,ExtName)) then
      evbadpix_tab = table(ev_set, ExtName)
      tablength = numberOfRows(evbadpix_tab)
      anothernumber=tablength+nbad
 
! Make arrays larger for mergeBad

      allocate(bad2x(2*anothernumber), bad2y(2*anothernumber),  &
               bad2type(2*anothernumber))
      allocate(bad2ext(2*anothernumber), bad2flag(2*anothernumber))
 
      if (.not. (overwrite .and. replaceonboard)) then
         call readBadpix(evbadpix_tab, .true., bad2x, bad2y, bad2ext, &
                                               bad2type, bad2flag, tablength)

! If overwrite, don't keep previous bad pixels except on-board ones
! If replaceonboard, don't keep on-board bright pixels

         if (overwrite .or. replaceonboard) then
            tablength2 = 0
            do k = 1,tablength
               if ((replaceonboard .and. bad2flag(k)==1) .or.  &
                   (overwrite) .and. bad2flag(k)>1)  cycle
               tablength2 = tablength2 + 1
               bad2x(tablength2)    = bad2x(k)
               bad2y(tablength2)    = bad2y(k)
               bad2ext(tablength2)  = bad2ext(k)
               bad2type(tablength2) = bad2type(k)
               bad2flag(tablength2) = bad2flag(k)
            enddo
            tablength = tablength2
         endif
      else
         tablength = 0
      endif

      newnumber = nbad + tablength
      bad2x(tablength+1:newnumber) = badx(1:nbad)
      bad2y(tablength+1:newnumber) = bady(1:nbad)
      bad2type(tablength+1:newnumber) = badtype(1:nbad)
      bad2ext(tablength+1:newnumber) = badyext(1:nbad)
      bad2flag(tablength+1:newnumber) = badflag(1:nbad)
      call mergeBad(bad2x, bad2y, bad2ext, bad2type, bad2flag, newnumber)
      call writeBadpix(evbadpix_tab, bad2x, bad2y, bad2ext, bad2type,  &
                       bad2flag, newnumber)

      deallocate(bad2x, bad2y, bad2ext, bad2type, bad2flag)
      call release(evbadpix_tab)
    else

!    If the bad pixel extension doesn't exist, create it and fill in the table

      call mergeBad(badx, bady, badyext, badtype, badflag, nbad)
      tab_bpix = addTable(ev_set,ExtName,nbad)
      badx_col = addcolumn(tab_bpix,"RAWX",Integer16,"pixels") 
      bady_col = addcolumn(tab_bpix,"RAWY",Integer16,"pixels") 
      badt_col = addcolumn(tab_bpix,"TYPE",Integer16,"") 
      bade_col = addcolumn(tab_bpix,"YEXTENT",Integer16,"pixels") 
      badf_col = addcolumn(tab_bpix,"BADFLAG",Integer16,"") 
      bad1x => Int16Data(badx_col)
      bad1y => Int16Data(bady_col)
      bad1type => Int16Data(badt_col)
      bad1ext => Int16Data(bade_col)
      bad1flag => Int16Data(badf_col)
      bad1x=badx(1:nbad)
      bad1y=bady(1:nbad)
      bad1ext=badyext(1:nbad)
      bad1type=badtype(1:nbad)
      bad1flag=badflag(1:nbad)

!	ADD A FEW ATTRIBUTES

      call setAttributes(attributable(tab_bpix), attributable(ev_set))
      call release(tab_bpix)
    endif 

  end subroutine addToBadPixExt

  subroutine loopoverallccds(NumBadPixExt, ev_set, ev_tab, instrument,  &
                             overwrite, replaceonboard, nbad)
  
!  Get the necessary information to update the bad pixel extension for each CCD

    integer,          intent(in)  :: NumBadPixExt
    type(DataSetT),   intent(in)  :: ev_set
    type(TableT),     intent(in)  :: ev_tab
    character(len=8), intent(in)  :: instrument
    logical,          intent(in)  :: overwrite, replaceonboard
    integer,          intent(out) :: nbad

    integer(kind=int16), allocatable, dimension(:) :: badx, bady, badyext
    integer(kind=int16), allocatable, dimension(:) :: badtype, badflag
    integer(int16), dimension(:), pointer :: UpX, UpY, UpType, UpYext
    logical, dimension(:), pointer :: UpUp
    integer             :: icount
    integer(kind=int32) :: node_id, ccd_id
    logical             :: ovwccd, obrepccd

    write (errstr,*) "Merging CCF bad pixels for CCDNR =",NumBadPixExt
    call message(AppMsg,Noisy)

    select case(trim(instrument))
    case("EMOS1", "EMOS2")
      node_id = NumBadPixExt/10
      ccd_id = NumBadPixExt - 10*node_id
      call CAL_setState(ccdChipId=ccd_id,ccdNodeId=node_id)  
    case default
      call CAL_setState(ccdChipId=NumBadPixExt)  
    end select

!   Get the bad pixels from the CCFs
    
    call CAL_getBadPixelList(UpX, UpY, UpType, UpYext, UpUp)
    nbad = size(UpX)
!    write (errstr,*) nbad, " bad pixels read in the CCF for CCDNR =", ccd_id
!  Changed this for now as the ccd_id is not yet defined for the pn, in 
!  response to the SSC-SPR-3441
    write (errstr,*) nbad, " bad pixels read in the CCF for CCDNR =", NumBadPixExt
    call message(AppMsg,Verbose)
    allocate(badx(2*nbad), bady(2*nbad), badyext(2*nbad))
    allocate(badtype(2*nbad), badflag(2*nbad))
    badx(1:nbad) = UpX; bady(1:nbad) = UpY
    badtype(1:nbad) = UpType; badyext(1:nbad) = UpYext

!   Give the badflag value for already known bad pixels (1) or CCF origin (2)

    flagforbadflaga: do icount=1,nbad
      if (UpUp(icount-1)) then
        badflag(icount) = 1
      else 
        badflag(icount) = 2
      end if
    end do flagforbadflaga
    call CAL_releaseMemory(UpX, UpY)
    call CAL_releaseMemory(UpType, UpYext); call CAL_releaseMemory(UpUp)

    ovwccd = overwrite
    obrepccd = replaceonboard
    if (evflagImage(ev_tab, NumBadPixExt, badx, bady, badyext,  &
                    badtype, badflag, nbad, ovwccd, obrepccd))  &
       call addToBadPixExt(ev_set, NumBadPixExt, ovwccd, obrepccd,  &
                           badx, bady, badyext, badtype, badflag, nbad)
    deallocate(badx, bady, badyext, badtype, badflag)

  end  subroutine loopoverallccds

! Add ':BADPIX' if name is that of a file

  subroutine testName(tableName)

! In/out: tablename : Name of the file/table

    character(len=*),  intent(inout)   :: tableName

    integer :: i

    i = index(tableName,':')
    if (i == 0)  tableName = trim(tableName) // ':BADPIX'

  end subroutine testName

end module ebadpixupdate_mod
